@@ -4,10 +4,8 @@ mod buffer;
 mod bytes;
 mod dict;
 mod list;
-mod pytype;
 
 pub use buffer::*;
 pub use bytes::*;
 pub use dict::*;
 pub use list::PyListIter;
-pub use pytype::*;
